#!/usr/bin/perl

#Copyright (C) 2006-2008 Keio University
#(Kris Popendorf) <comp@bio.keio.ac.jp> (2006)
#
#This file is part of Murasaki.
#
#Murasaki is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#Murasaki is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with Murasaki.  If not, see <http://www.gnu.org/licenses/>.

use strict;
use File::Basename;

BEGIN {
  our $root=(fileparse($0))[1];
  $root=$ENV{MURASAKI} if $ENV{MURASAKI};
  unshift(@INC,$root);
  require "$root/config.pl";
  require "$root/common.pl";
  our $seqhome="$root/seq";
  unshift(@INC,$root);
}

use Getopt::Long;
use Pod::Usage;
$Term::ANSIColor::AUTORESET=1;

#use Data::Dump qw{dump};
use Ticker qw{resetTick tick};

our $root;

our %units=(gb => 1024*1024*1024, mb => 1024*1024, kb=>1024, b=>1, tb=>1024*1024*1024*1024);
my $geneparse="$root/geneparse";

my ($help,$man,$verbose);
my $err=GetOptions('help|?' => \$help, man => \$man, 'verbose+'=>\$verbose,
		  );
pod2usage(1) if $help or !$err or @ARGV<1;
pod2usage(-exitstatus => 0, -verbose => 4) if $man;

foreach my $file (@ARGV){
  local $\="\n";
  my $otf="$file.len";
  print STDERR "Writing out to $otf" if $verbose;
  open(my $otfh,">$otf") or die "Couldn't write to $otf";
  my $length=`$geneparse -c -l $file`;
  print $otfh $length;
  my $type=`$geneparse -t -c $file`;
  if($type eq 'stitch'){
    my @subfiles=readStitchFile($file);
    my $start=0;
    foreach my $subfile (@subfiles){
      print $otfh join("\t",$subfile->{len},(fileparse($subfile->{file}))[0]);
    }
  }elsif($type eq 'fasta'){
    open(my $fh,$file) or die "Couldn't open $file";
    my ($count,$countall,$name)=(0,0,undef);
    while(<$fh>){
      chomp;
      next if m/^;/; #fasta comments
      if(m/^>(.*)$/){
	if(defined $name){
	  print $otfh join("\t",$count,$name);
	  $countall+=$count;
	  $name=$1;
	  $countall+=10;
	  $count=0;
	}else{
	  $name=$1;
	}
      }else{
	s/[^ATGCN]//gi;
	$count+=length($_);
      }
    }
    print $otfh join("\t",$count,$name);
    $countall+=$count;
    warn "Expected $length, but read $countall" unless($countall==$length);
  }else{
    print $otfh join("\t",$length,(fileparse($file))[0]);
  }
}

sub totalLen {
  my @files=@_;
  return sum(map {$_->{len}} @files);
}

sub sum {
  my $r=0;
  foreach my $i (@_){
    $r+=$i;
  }
  return $r;
}

sub max {
  my ($r,@l)=@_;
  foreach my $v (@l){
    $r=$v if $v>$r;
  }
  return $r;
}

sub makeIds {
  my ($ref,@others)=@_;
  my $re=qr/\W+/;
  my @bits=split($re,$ref);
  my @use=map {0} 0..$#bits;
  foreach my $other (@others){
    my @obits=split($re,$other);
    for my $i (0..max($#obits,$#bits)){
      $use[$i]=1 if $bits[$i] ne $obits[$i];
    }
  }
  my @ids;
  my @slice=toSlice(@use);
  @slice=(0) unless @slice;
  foreach my $name ($ref,@others){
    my @bits=split($re,$name);
    push(@ids,join(".",@bits[@slice]));
  }
  return @ids;
}

sub humanMemory {
  my @l=@_;
  @l=map {
    my $minus=($_<0);
    $_=abs($_);
    my $scale=$_/$units{b}<500 ? "b":
      $_/$units{kb}<500 ? "kb":
	$_/$units{mb}<500 ? "mb":
	  $_/$units{gb}<500 ? "gb":"tb";
    my $str=sprintf("%.3f",$_/$units{$scale});
    $str=~s/(\.\d+?)0+$/$1/;
    $str=~s/\.$//;
    (($minus ? "-":"").$str.$scale)} @l;
  return @l if wantarray;
  return $l[0];
}

sub readStitchFile {
  my ($file)=@_;
  my @files;
  open(my $fh,$file) or die "Couldn't open $file";
  while(my $line=<$fh>){
    chomp $line;
    my ($file,$len,$start,$stop)=split(/\t/,$line);
    die "Not a valid stitch file" unless ($file and $len and $start and $stop);
    my %dat;
    @dat{qw{file len start stop}}=($file,$len,$start,$stop);
    push(@files,\%dat);
  }
  return @files;
}

###stuff for project

sub findRegions {
  my @seqs=@_;
  my %regions;
  foreach my $si (0..$#seqs){
    my $seq=$seqs[$si];
    my ($name,$dir,$ext)=fileparse($seq,qr/\.[^\.]+/);
    if($ext eq '.stitch'){
      %regions=(%regions,findStitchRegions($seq,$si));
    }else{
      $regions{$seq}={seq=>$seq,si=>$si,start=>1};
#      print "si $si : Raw region: ".regionStr($regions{$seq})."\n";
    }
  }
  return %regions;
}

sub findStitchRegions {
  my ($stitch,$si,$start)=@_;
  $start=1 unless $start;
  my @files=readStitchFile($stitch);
  my %regions;
  foreach my $file (@files){
    my ($name,$dir,$ext)=fileparse($file->{file},qr/\.[^\.]+/);
    if($ext eq '.stitch'){
      %regions=(%regions,findStitchRegions($file->{file},$si,$file->{start}+$start-1));
    }else{
      $regions{$file->{file}}={seq=>$file->{file},si=>$si,start=>$start+$file->{start}-1};
#      print "si $si : Stitch region: ".regionStr($regions{$file->{file}})."\n";
    }
  }
  return %regions;
}

sub regionStr {
  return join(",",map { "(".join(",",@{$_}{qw{seq si start}}).")"} @_);
}


__END__

=head1 NAME

getsegments.pl -- Generate segment specific information for multi-segment files

=head1 SYNOPSIS

getsegments.pl file1.fa file2.stitch
#generates statistics for file1 and file2
getsegments.pl file3.gbk
#genebank doesn't use multiple segments, so this will diligently produce a .len, but with only 1 entry

=head1 OPTIONS

--verbose => makes more verbose
